package adt.mainnode.config;

import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import springfox.documentation.builders.RequestHandlerSelectors;
import springfox.documentation.service.ApiInfo;
import springfox.documentation.service.Contact;
import springfox.documentation.service.Tag;
import springfox.documentation.spi.DocumentationType;
import springfox.documentation.spring.web.plugins.Docket;
import springfox.documentation.swagger2.annotations.EnableSwagger2;

import java.util.Collections;

import static springfox.documentation.builders.PathSelectors.regex;

@EnableSwagger2
@Configuration
public class SwaggerConfig {

    @Bean
    public Docket DocketApi() {
        return new Docket(DocumentationType.SWAGGER_2)
                .groupName("ADT Main Node")
                .select()
                .apis(RequestHandlerSelectors.any())
                .paths(regex("/(vlans|configurations|nlp|virt_query|event|presence_check).*"))
                .build()
                .apiInfo(metaInfo())
                .tags(
                        new Tag("Vlans", "Endpoints to create new assets or to retrieve existing ones."),
                        new Tag("Configurations", "Endpoint queried by distributed nodes the first time they start up. Grants them a scanning configuration.")  ,
                        new Tag("NLP", "Natural Language Processing functionalities."),
                        new Tag("Virt_Query", "Endpoint to query the Virtuoso Database."),
                        new Tag("(Event) Transfer Service", "Endpoint to send vulnerabilities' events to a pre-set URL.")
                );
    }

    private ApiInfo metaInfo() {
        return new ApiInfo(
                "ADT Main Node",
                "This is the online documentation of ADT Main Node. Click authorize to perform queries with your own credentials.",
                "ALPHA 2",
                "Terms of service",
                new Contact("Roberto Garrido",
                        "https://medal.ctb.upm.es/roberto-garrido-garcia",
                        "roberto.garrido@ctb.upm.es"),
                "License",
                "License URL",
                Collections.emptyList()
        );
    }

}
