package adt.mainnode.controller;

import adt.mainnode.entity.Credentials;
import adt.mainnode.entity.Notification;
import adt.mainnode.repository.CredentialsRepository;
import adt.mainnode.service.NotificationService;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.*;

import javax.servlet.http.HttpServletResponse;

@RestController
@RequestMapping("notifications")
@Api(
        value = "/notifications",//"/asset",
        produces = MediaType.APPLICATION_JSON_VALUE,
        tags = {"Notifications"}
)
@Slf4j
public class NotificationController {

    private final CredentialsRepository credentialsRepository;
    private final NotificationService notificationService;

    @Autowired
    public NotificationController(CredentialsRepository credentialsRepository, NotificationService notificationService){
        this.credentialsRepository = credentialsRepository;
        this.notificationService = notificationService;
    }

    @CrossOrigin
    @ApiOperation(value = "Save a new set of notification", response = Boolean.class)
    @ResponseStatus(HttpStatus.CREATED)
    @PostMapping(produces = MediaType.APPLICATION_JSON_VALUE)
    boolean saveAssets(HttpServletResponse response, @RequestBody Notification notification, @RequestParam String token) {
        Credentials credential = credentialsRepository.findByToken(token);
        if(credential != null && credential.getStatus().equals("OPERATIVE")) {

            if(credential.getAvailableEndpoint().contains("(POST) notifications/")) {
                log.info("New notification received");
                notificationService.saveNotification(notification, credential);
                return true;
                }
        }
       return false;
    }
}

