package adt.mainnode.service;

import adt.mainnode.entity.Node;
import adt.mainnode.entity.ScanConfig;
import adt.mainnode.entity.TPMHash;
import adt.mainnode.entity.Vlan;
import adt.mainnode.identity.VlanIdentity;
import adt.mainnode.repository.TPMHashRepository;
import adt.mainnode.repository.NodeRepository;
import adt.mainnode.repository.ScanConfigRepository;
import adt.mainnode.repository.VlanRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import lombok.extern.slf4j.Slf4j;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;

@Slf4j
@Service
public class InitializeService {

    private final NodeRepository nodeRepository;
    private final VlanRepository vlanRepository;
    private final ScanConfigRepository scanConfigRepository;
    private final TPMHashRepository tpmHashRepository;
    private String serverIP;
    private String serverPort;
    private String tpmPresence;

    private String transferUser;
    private String transferPassword;
    private String petitionUrl;

    private String transferUrl;

    private String transferToken;
    private Date transferTokenReceipt;

    private TPMHash tpmHash;
    private String pcrHash;


    @Autowired
    public InitializeService(NodeRepository nodeRepository, TPMHashRepository tpmHashRepository, ScanConfigRepository scanConfigRepository, VlanRepository vlanRepository){
        this.nodeRepository = nodeRepository;
        this.scanConfigRepository = scanConfigRepository;
        this.vlanRepository = vlanRepository;
        this.tpmHashRepository = tpmHashRepository;
        this.transferToken = null;
        this.serverIP = null;
        this.serverPort = null;
        this.tpmPresence = null;
        this.pcrHash = null;

        if(tpmHashRepository.findAll().size() > 0) {
            this.tpmHash = tpmHashRepository.findAll().get(0);
        }else{
            log.info("**************___sero");
            this.tpmHash = null;
        }
        log.info("**************___{}",tpmHash);

    }


    public String getServerIP() {
        return serverIP;
    }

    public void setServerIP(String serverIP) {
        this.serverIP = serverIP;
    }


    public String getServerPort() {
        return serverPort;
    }

    public void setServerPort(String serverPort) {
        this.serverPort = serverPort;
    }

    public String getTpmPresence() {
        return tpmPresence;
    }

    public void setTpmPresence(String tpmPresence) {
        this.tpmPresence = tpmPresence;
    }


    public String getTransferUser() {
        return transferUser;
    }

    public void setTransferUser(String transferUser) {
        this.transferUser = transferUser;
    }


    public String getTransferPassword() {
        return transferPassword;
    }

    public void setTransferPassword(String transferPassword) {
        this.transferPassword = transferPassword;
    }


    public String getPetitionUrl() {
        return petitionUrl;
    }

    public void setPetitionUrl(String petitionUrl) {
        this.petitionUrl = petitionUrl;
    }


    public String getTransferUrl() {
        return transferUrl;
    }

    public void setTransferUrl(String transferUrl) {
        this.transferUrl = transferUrl;
    }


    public String getTransferToken() {
        return transferToken;
    }

    public void setTransferToken(String transferToken) {
        this.transferToken = transferToken;
    }


    public Date getTransferTokenReceipt() {
        return transferTokenReceipt;
    }

    public void setTransferTokenReceipt(Date transferTokenReceipt) { this.transferTokenReceipt = transferTokenReceipt; }


    public TPMHash getTPMHash() {
        return tpmHash;
    }

    public void setTPMHash(TPMHash tpmHash) {
        this.tpmHash = tpmHash;
    }


    public String getPCRHash() {
        return pcrHash;
    }

    public void setPCRHash(String pcrHash) {
        this.pcrHash = pcrHash;
    }


    // TODO: Obtener la gateway
    public Node saveNode(Node node, List<ScanConfig> scanConfigs, boolean flag) {

        List<Vlan> vlans = new ArrayList<>();

        for(int i=0; i<scanConfigs.size(); i++) {

            ScanConfig scanConfig = scanConfigs.get(i);

            VlanIdentity vlanId = VlanIdentity.createVlanIdentity(node.getNodeId(), scanConfig.getVlanName());

            Vlan vlan = Vlan.createVlan(vlanId, scanConfig.getDelay(), scanConfig.getIpToScan(), scanConfig.getIpToScan(), scanConfig.getMask());//"192.168.1.0",null, "24");
            vlans.add(vlan);
        }
        //node.setVlan(vlan);
        node.setVlan(vlans);

        vlanRepository.saveAll(vlans);

        //if(!flag){
        nodeRepository.save(node);
        //}

        return node;//nodeRepository.save(node);
    }


}
